clear;

% Analysis of temporal integration in melody discrimination using psychophysical kernels/templates 
% (see Neri & Heeger, Nat Neuro 2002)
%
% the data are coded as 
% data{sub_id}.(trial, :)  with the following elements [RT,responsekey_id,Correct/wrong,motion direction ,coherence,coherence threshold,rnd seedval];
%          key_id 38 =up, 40 =down; Correct 1 / wrong 0;  direction 1 = up, 2= down;  coherence: fractional motion coherence [0 to 1]
%
% MOTLIST{sub_id}(trial,:) codes the moment by moment motion evidence in the stimulus, coded in 
% time bins (frames), whose length and number changes across datasets
%
% The script uses the VBA Toolbox https://mbb-team.github.io/VBA-toolbox/
% and the Bayesian regression package https://de.mathworks.com/matlabcentral/fileexchange/60823-bayesian-penalized-regression-with-continuous-shrinkage-prio
%


if 1==0
  addpath('D:\CKDATA\MatBox\toolboxes\VBA-toolbox-master')
  addpath('D:\CKDATA\MatBox\toolboxes\VBA-toolbox-master\modules\GLM')
  run('D:\CKDATA\MatBox\toolboxes\VBA-toolbox-master\VBA_setup.m')
  addpath('D:\CKDATA\MatBox\toolboxes\Breg')
end

SET =4 % Experiment number

DATAPATH = './';

ARG.flist = [1:0.1:2,2.2:0.2:5]; % frequencies used for fitting frequency dependent models. These need to be above the duration of the stimulus

% Number of randomizations for bootstrapping kernels, boostrap CI's. Affects run-time considerably.
ARG.nboot = 4000; % set to 200 for testing, to 4000 for final analysis.
ARG.Freqsmooth = 5; % width of Gaussian window for spectral smoothing in FFT analysis. Not that important
ARG.version = 'V6'; 

KeyResult.ARG = ARG;

%--------------------------------------------------------------------------------
% load data 
%--------------------------------------------------------------------------------
sname = sprintf('%sAlldata_version%d.mat',DATAPATH,SET);
load(sname,'MOTLIST','data');
ntime =  size(MOTLIST{1}(1,:),2);
time_ax = [1:ntime]*0.03; % time axis in sec
nsub = length(data);

%--------------------------------------------------------------------------------
% compute kernels for each participant using bootstrapping to render units comparable as in Neri et al
%--------------------------------------------------------------------------------
KeyResult.Basics.NSub  = length(data);
AllKernels=[];
for S=1:length(data)
  KeyResult.Basics.Ntrials(S) = size(data{S},1);
  KeyResult.Basics.Behav_Thres(S) = mean(data{S}(:,6));
  % compute response template for condition 1 (arbitrary)
  hit = find( (data{S}(:,2)==38).*(data{S}(:,4)==1));
  fa = find( (data{S}(:,2)==38).*(data{S}(:,4)==2));
  cr = find( (data{S}(:,2)==40).*(data{S}(:,4)==2));
  miss = find( (data{S}(:,2)==40).*(data{S}(:,4)==1));
  % z-score  this makes the result non-deterministic !
  [~,kernel] = ck_psych_RespTemplate(MOTLIST{S}(:,[1:ntime]),hit,fa,miss,cr,ARG.nboot);
  AllKernels(S,:) = kernel;
end
KeyResult.PercepWeight = AllKernels;
KeyResult.time_ax = time_ax;

tmp = zeros(ARG.nboot,ntime);
for b=1:ARG.nboot
  r = ceil(rand(1,nsub)*nsub);
  tmp(b,:) = mean(AllKernels(r,:));
end
KeyResult.PercepWeight_CI = prctile(tmp,[2.5,97.5]);

%--------------------------------------------------------------------------------
% - Power spectrum with group-level CI 
%--------------------------------------------------------------------------------
GW = gausswin(ARG.Freqsmooth); GW = GW/sum(GW);
Fitpower=[];
for S=1:nsub
  % standardize overall power
  tmp = detrend(AllKernels(S,:));
  tmp = tmp-mean(tmp(:));
  tmp = tmp./std(tmp(:));
  % frequency spectrum after zero-padding to fixed common resolution across paradigms
  L2 = ceil((180-ntime)/2); 
  [p,fax] = ck_powerspec([zeros(1,L2),tmp,zeros(1,L2)],33.3);
  fax =fax(2:end); p = p(2:end);
  %  smoothen with Gaussian window
  Fitpower(S,:) = ck_filt_conv(log(p),GW);
end
KeyResult.Spectrum.power = Fitpower;
KeyResult.Spectrum.fax = fax;

% group-level CI
tmp = zeros(ARG.nboot,length(fax));
for b=1:ARG.nboot
  r = ceil(rand(1,nsub)*nsub);
  tmp(b,:) = mean(Fitpower(r,:));
end
KeyResult.Spectrum.power_CI = prctile(tmp,[0.5 99.5]);
%  peak of group-level mean
[~,fp] = max(mean(Fitpower),[],2);
KeyResult.Spectrum.peak = fax(fp);

% null model spectrum  based on time shuffling 
% tests the hypothesis of power being higher than expected from 
% white noise at this binning
tmpB = zeros(ARG.nboot,length(fax));
FitpowerNull=[];
for b=1:ARG.nboot
  rand_order = randperm(ntime);
  for S=1:nsub
    tmp = detrend(AllKernels(S,:));
    tmp = tmp-mean(tmp(:));
    tmp = tmp./std(tmp(:));
    tmp = tmp(rand_order); % shuffle
    [p,f] = ck_powerspec([zeros(1,L2),tmp,zeros(1,L2)],33.3);
    p = p(2:end);
    FitpowerNull(S,:) = ck_filt_conv(log(p),GW);
   
  end
  r = ceil(rand(1,nsub)*nsub);
  tmpB(b,:) = mean(FitpowerNull(r,:));
end
KeyResult.Spectrum.Rand_CI = prctile(tmpB,[2.5,97.5]);


%% --------------------------------------------------------------------------------
% Model comparison of 'trivial models'. 
% here we ask how well the templates are fit by a 
% - constant offset (should be there by design, performance around 75% correct)
% - constant + linear slope,
% - constant + slope +  a U/V shaped term adapted to the stimulus duration
% for each participant we compute the log evidence from three regression models
%--------------------------------------------------------------------------------
LogEv=[];
for S=1:nsub
  y = AllKernels(S,:)';
  lin = [1:length(y)]/length(y);
  xs  = 2*pi*[1:ntime]*0.03;
  
  % 1) constant model 
  X = ones(length(y),1);
  b = X\y;
  LogEv(S,1) = lev_GLM(y,X);
  
  % 2) linear model
  X = ones(length(y),2);
  X(:,2) = lin;
  b = X\y;
  LogEv(S,2) = lev_GLM(y,X);
  
  % 3) 'cosine' model with u/v shaped weights dependent on stimulus duration
  X = ones(length(y),3);
  X(:,2) = lin;
  X(:,3) = cos(xs*(1/time_ax(end)));
  b = X\y;
  LogEv(S,3) = lev_GLM(y,X);
end

% Subject-specific AICc 
% AICc = AIC + 2*k*(k+1) / (n-k-1)
n = ntime; 
AICc=[];
k=1;
AICc(:,1) = -2*LogEv(:,1) + 2*k + 2*k*(k+1)/(n-k-1);
k=2;
AICc(:,2) = -2*LogEv(:,2) + 2*k + 2*k*(k+1)/(n-k-1);
k=3;
AICc(:,3) = -2*LogEv(:,3) + 2*k + 2*k*(k+1)/(n-k-1);

% group-level VBA
options.verbose = 0;
options.DisplayWin = 0;
[p1, vba] = VBA_groupBMC (-0.5*AICc',options);
KeyResult.trivial.EP = vba.ep;
KeyResult.trivial.EMF = vba.Ef';

% group-level BIC values
AICcg = sum(AICc,1);
KeyResult.trivial.AICcg = AICcg;

% determine best trivial model 
[~,KeyResult.best_triv] = min(AICcg);


%% --------------------------------------------------------------------------------
% Model comparison of the best trivial model and rhythmic models at different frequencies
% we fit both models using 2-fold cross-validation, repeated x times
% here we need to recompute the templates for the CV loop (slow)
%--------------------------------------------------------------------------------
clear AIC* BIC*
clear LogEv*

for S=1:length(data)
  cvrun=1;
  for R=1:10 % repeated runs of 2-fold CV
    cv = cvpartition(size(data{S}(:,2),1), 'kfold',2);
    for c=1:2
      trainIdx = cv.training(c);
      testIdx = cv.test(c);

      % first fold
      hit = find( (data{S}(trainIdx,2)==38).*(data{S}(trainIdx,4)==1));
      fa = find( (data{S}(trainIdx,2)==38).*(data{S}(trainIdx,4)==2));
      cr = find( (data{S}(trainIdx,2)==40).*(data{S}(trainIdx,4)==2));
      miss = find( (data{S}(trainIdx,2)==40).*(data{S}(trainIdx,4)==1));
      [~,templateZ_1] = ck_psych_RespTemplate(MOTLIST{S}(trainIdx,[1:ntime]),hit,fa,miss,cr,ARG.nboot);
      
      % 2nd fold
      hit = find( (data{S}(testIdx,2)==38).*(data{S}(testIdx,4)==1));
      fa = find( (data{S}(testIdx,2)==38).*(data{S}(testIdx,4)==2));
      cr = find( (data{S}(testIdx,2)==40).*(data{S}(testIdx,4)==2));
      miss = find( (data{S}(testIdx,2)==40).*(data{S}(testIdx,4)==1));
      [~,templateZ_2] = ck_psych_RespTemplate(MOTLIST{S}(testIdx,[1:ntime]),hit,fa,miss,cr,ARG.nboot);
      
      %--------------------------------------------------------------------------------
      % Model comparison of the best trivial model and rhythmic models at different frequencies
      y = templateZ_1';
      lin = [1:length(y)]/length(y);
      xs  = 2*pi*[1:ntime]*0.03;
      
      % best trivial model
      X = ones(length(y),3);
      X(:,2) = lin;
      X(:,3) = cos(xs*(1/time_ax(end)));
      X = X(:,[1:KeyResult.best_triv]); % reduce depending on above result
      b = X\y;
      % compute log evidence on fold 2
      LogEv_1 = lev_GLM(templateZ_2',X);
      
      % 2) rhythmic models at different frequencies
      LogEv_2 =[];
      for f=1:length(ARG.flist)
        X2 = ones(length(y),2+KeyResult.best_triv);
        X2(:,[1:KeyResult.best_triv]) = X;
        X2(:,KeyResult.best_triv+1) = cos(xs*ARG.flist(f));
        X2(:,KeyResult.best_triv+2) = sin(xs*ARG.flist(f));
        b = X2\y;
        LogEv_2(f,:) = lev_GLM(templateZ_2',X2);
      end
      % AICc
      n = ntime; k =  KeyResult.best_triv;
      AICc(S,cvrun) = -2*LogEv_1 + 2*k + 2*k*(k+1)/(n-k-1);
      n = ntime; k = KeyResult.best_triv+2;
      AICc_f(S,cvrun,:) = -2*LogEv_2 + 2*k + 2*k*(k+1)/(n-k-1);
      % BIC 
      BIC(S,cvrun) = -2*LogEv_1 + KeyResult.best_triv*log(ntime);
      BIC_f(S,cvrun,:) =  -2*LogEv_2 + (KeyResult.best_triv+2)*log(ntime);
     
      cvrun =cvrun+1;
    end
  end
end

% average folds
AICc = mean(AICc,2);
AICc_f = squeeze(mean(AICc_f,2));
BIC = mean(BIC,2);
BIC_f = squeeze(mean(BIC_f,2));

% find each subject's specific best frequency
[~,f_best_each] = min(AICc_f,[],2);
KeyResult.Freqmod.Bestfreq_each_AICc = ARG.flist(f_best_each);

% group-level AICc
AICcg = sum(AICc);
AICc_fg = sum(AICc_f);
KeyResult.Freqmod.AICcg_f = [AICcg AICc_fg];

BICg = sum(BIC);
BIC_fg = sum(BIC_f);
KeyResult.Freqmod.BICg_f = [BICg BIC_fg];

% find best group-level frequency
[~,fbestaic] = min(AICc_fg);
KeyResult.Freqmod.BestfreqG_AICc = ARG.flist(fbestaic);


%% --------------------------------------------------------------------------------
% Model comparison between  best trivial model and the best rhythmic model based on the 
% group-level frequency
%--------------------------------------------------------------------------------
AICc2 = cat(2,AICc,AICc_f(:,fbestaic));
[p1, vba_aic] = VBA_groupBMC (-0.5*AICc2',options);
KeyResult.Freqmod.AIC_EP = vba_aic.ep;
KeyResult.Freqmod.AIC_EMF = vba_aic.Ef';

KeyResult.Freqmod.AICcg_modelcomp = sum(AICc2);

BICc2 = cat(2,BIC,BIC_f(:,fbestaic));
[p1, vba_aic] = VBA_groupBMC (-0.5*BICc2',options);
KeyResult.Freqmod.BIC_EP = vba_aic.ep;
KeyResult.Freqmod.BIC_EMF = vba_aic.Ef';

KeyResult.Freqmod.BICcg_modelcomp = sum(BICc2);


%% ---------------------------------------------------------------------------
% analysis of winning model
% here we recompute the model weights without CV on the full data 
% to collect beta's and best phase  values
% best trivial model + best rhythmic model at group-level frequency
%-----------------------------------------------------------------------------

lin = [1:ntime]/ntime;
xs  = 2*pi*[1:ntime]*0.03;

X = ones(ntime,3);
X(:,2) = lin;
X(:,3) = cos(xs*(1/time_ax(end)));
X = X(:,[1:KeyResult.best_triv]); % reduce depending on above result
X2 = ones(length(y),2+KeyResult.best_triv);
X2(:,[1:KeyResult.best_triv]) = X;
X2(:,KeyResult.best_triv+1) = cos(xs*KeyResult.Freqmod.BestfreqG_AICc );
X2(:,KeyResult.best_triv+2) = sin(xs*KeyResult.Freqmod.BestfreqG_AICc );

% single subject betas
for S=1:nsub
  y = AllKernels(S,:)';
  b = X2\y;
  Beta_w(S,:) = b;
end
KeyResult.Freqmod.Beta_w = Beta_w;

% display group-level average model 
mw = mean(Beta_w);
KeyResult.Freqmod.GroupModel = X2*mw';


% display each subjects rhythmic component 
for S=1:nsub
  KeyResult.Freqmod.EachModel(S,:) = X2(:,[-1,0]+end)*Beta_w(S,[-1,0]+end)';
end



%% --------------------------------------------------------------------------------
% group-level analysis based on Monte-Carlo simulations (using the Bayesian regression toolbox)
% and the WAIC weight to estimate the out-of-sample variance

WAIC=[];
for S=1:length(data)
  
  y = AllKernels(S,:)';
  lin = [1:length(y)]/length(y);
  xs  = 2*pi*[1:ntime]*0.03;
  
  % best trivial model
  X = ones(length(y),2);
  X(:,1) = lin;
  X(:,2) = cos(xs*(1/time_ax(end)));
  [b, b0, retval] = bayesreg(X,y,'gaussian','g','nsamples',1e4,'burnin',1e4,'thin',5,'display',false);
  WAIC(S,1) =retval.modelstats.waic;
  
  % 2) rhythmic models at different frequencies
  for f=1:length(ARG.flist)
    X2 = ones(length(y),4);
    X2(:,[1:2]) = X;
    X2(:,2+1) = cos(xs*ARG.flist(f));
    X2(:,2+2) = sin(xs*ARG.flist(f));
    [b, b0, retval] = bayesreg(X2,y,'gaussian','g','nsamples',1e4,'burnin',1e4,'thin',5,'display',false);
    WAIC(S,f+1) =retval.modelstats.waic;
  end
  
end
KeyResult.Freqmod.WAIC = WAIC;


sname = sprintf('%sResults_V6_%d.mat',DATAPATH,SET);
save(sname,'KeyResult')


return;
